@@ -20,6 +20,7 @@ from colorama import Style
 from colorama import init as colorinit
 import ete3
 from ansi2html import Ansi2HTMLConverter
+import multiprocessing as mp
 import logging
 import pathlib
 import gc
@@ -27,19 +28,27 @@ import gc
 # from beautifultable import BeautifulTable
 
 #### Get Version
+__version__ = "unspec"
 try:
     from importlib.metadata import version, PackageNotFoundError
 
     __version__ = version(__name__)
 except:
-    # if module or this python package is not installed
-    from pkg_resources import get_distribution, DistributionNotFound
-
     try:
-        __version__ = get_distribution(__name__).version
-    except DistributionNotFound:
-        # package is not installed
-        __version__ = "unspec"
+        # if the python version is <3.8 this module is used as backport
+        from importlib_metadata import version, PackageNotFoundError
+
+        __version__ = version(__name__)
+    except:
+        # if module or this python package is not installed
+        from pkg_resources import get_distribution, DistributionNotFound
+
+        try:
+            __version__ = get_distribution(__name__).version
+        # except DistributionNotFound:
+        except:
+            # package is not installed
+            __version__ = "unspec"
 ####
 
 
@@ -62,10 +71,14 @@ def writeable_dir(prospective_dir):
 def parse_args():
     parser = argparse.ArgumentParser(
         prog="Vicinator",
-        description="Track Microsynteny of a target proteins and its orthologs across genomes",
+        description="Track Microsynteny of target proteins and its orthologs across genomes.",
     )
 
-    parser.add_argument(
+    parser._action_groups.pop()
+
+    required = parser.add_argument_group('required arguments')
+
+    required.add_argument(
         "--tabular-ortholog-groups",
         dest="ogtable",
         metavar="<orthology_table>",
@@ -74,62 +87,7 @@ def parse_args():
         help="path to mapping file with format ortholog_group_id<tab>genome_id<tab>protein_seq_id",
     )
 
-    parser.add_argument(
-        "--reference",
-        dest="ref_feat_table",
-        metavar="<reference_genome_feature_table_file_path>",
-        type=argparse.FileType("rt"),
-        required=True,
-        help="path to a ncbi style feature table file that acts as a reference",
-    )
-
-    parser.add_argument(
-        "--tree",
-        dest="tree",
-        metavar="<newick_tree_file_path>",
-        type=argparse.FileType("rt"),
-        required=False,
-        help="path to newick tree that includes all taxa to be screened",
-    )
-
-    #TODO: FUTURE_FEATURE
-    # parser.add_argument(
-    #     "--intergenic-distances",
-    #     dest="genedist",  # metavar='<newick_tree_file_path>',
-    #     action="store_true",
-    #     required=False,
-    #     help="if option is set, shows intergenic distances of genes surrounding the center gene",
-    # )
-
-    parser.add_argument(
-        "--force",
-        dest="force_new_database",  # metavar='<newick_tree_file_path>',
-        action="store_true",
-        required=False,
-        default=False,
-        help="if option is set, existing ortholog databases in the output dir are ignored and will be overwritten",
-    )
-
-    parser.add_argument(
-        "--prefix",
-        dest="output_prefix",  # metavar='<newick_tree_file_path>',
-        action="store_true",
-        type=str,
-        metavar="<prefix_str_for_output>",
-        required=False,
-        help="if option is set, shows intergenic distances of genes surrounding the center gene",
-    )
-
-    parser.add_argument(
-        "--centerprotein-accession",
-        dest="centerprotein_accession",
-        metavar="<protein_accession_str>",
-        type=str,
-        required=True,
-        help="unique identifier of the central gene of the window",
-    )
-
-    parser.add_argument(
+    required.add_argument(
         "--feat-tables-dir",
         dest="feat_tables_dir",
         metavar="<dir_path>",
@@ -138,7 +96,36 @@ def parse_args():
         help="path to directory of *.feature_tables.txt or *.gff3 files that shall be screen",
     )
 
-    parser.add_argument(
+    # TODO: FUTURE_FEATURE
+    # parser.add_argument(
+    #     "--intergenic-distances",
+    #     dest="genedist",  # metavar='<newick_tree_file_path>',
+    #     action="store_true",
+    #     required=False,
+    #     help="if option is set, shows intergenic distances of genes surrounding the center gene",
+    # )
+
+    n_group = parser.add_argument_group('required arguments (neighborhood)')
+
+    n_group.add_argument(
+        "--reference",
+        dest="ref_feat_table",
+        metavar="<file_path>",
+        type=argparse.FileType("rt"),
+        required=True,
+        help="path to a ncbi style feature table or gff file that acts as a reference",
+    )
+
+    n_group.add_argument(
+        "--centerprotein-accession",
+        dest="centerprotein_accession",
+        metavar="<str>",
+        type=str,
+        required=True,
+        help="unique identifier of the central gene of the window",
+    )
+
+    n_group.add_argument(
         "--extension-size",
         dest="k",
         metavar="<int>",
@@ -147,22 +134,63 @@ def parse_args():
         help="defines the #features that are co-checked to the left and right of the centerprotein",
     ),
 
-    parser.add_argument(
+    o_group = parser.add_argument_group("optional arguments (output)")
+
+    o_group.add_argument(
+        "--tree",
+        dest="tree",
+        metavar="<newick_tree_file_path>",
+        type=argparse.FileType("rt"),
+        required=False,
+        help="path to newick tree that includes all taxa to be screened",
+    )
+
+    o_group.add_argument(
         "--outdir",
         dest="outdir",
         metavar="<dir_path>",
         type=writeable_dir,
-        required=True,
+        required=False,
+        default=os.getcwd(),
         help="path to desired output directory",
     )
 
-    parser.add_argument(
+    o_group.add_argument(
+        "--prefix",
+        dest="output_prefix",  # metavar='<newick_tree_file_path>',
+        type=str,
+        metavar="<str>",
+        required=False,
+        help="if option is set, shows intergenic distances of genes surrounding the center gene",
+    )
+
+    o_group.add_argument(
         "--outputlabel-map",
         dest="labelmap",
         metavar="<file_path>",
         type=argparse.FileType("rt"),
         default=None,
-        help="Optional mapping file feature file accession <tabs> replacement string",
+        help="Attempts to replace genome accessions in the outputs with a replacement string. Requires a two-column map file formatted like so: 'genome file accession' <tab> 'replacement string'",
+    )
+
+    r_group = parser.add_argument_group('optional arguments (run)')
+
+    r_group.add_argument(
+        "--nprocs",
+        dest="nprocs",
+        metavar="<int>",
+        type=int,
+        default=mp.cpu_count()-1,
+        help="Number of CPUs for parallel processing of genomes. Default: Number of CPUs-1",
+    )
+
+    r_group.add_argument(
+        "--force",
+        dest="force_new_database",  # metavar='<newick_tree_file_path>',
+        action="store_true",
+        required=False,
+        default=False,
+        help="if option is set, existing ortholog databases in the output dir are ignored and will be overwritten",
     )
 
     parser.add_argument("--version", action="version", version=__version__)
@@ -222,9 +250,6 @@ class Genome:
 
             ft = ft.drop(["start.old", "end.old", "strand.old"], axis=1)
 
-            # TODO: Major issue here: When reversing the df, CDS features don't succeed their gene feature annot any longer
-            # TODO: An idea: first  place all gene features and then fill from top the inbetween spaces
-
             return ft
 
         if len(self.feature_df["OG"]):
@@ -378,6 +403,7 @@ class Genome:
 
         upstream = []
         downstream = []
+
         for i in range(1, k + 1):
             if c - i >= 0:
                 upstream.append(df.iloc[c - i].name)
@@ -447,8 +473,10 @@ class Genome:
 
 
 def readOGTable(orthotable_filepath, outdir, force_new_database):
-    if not os.path.exists(os.path.join(outdir, "vicinator.ogtable.pickle")) or \
-            force_new_database:
+    if (
+        not os.path.exists(os.path.join(outdir, "vicinator.ogtable.pickle"))
+        or force_new_database
+    ):
         ogtable = pd.read_csv(
             orthotable_filepath,
             sep="\t",
@@ -611,15 +639,18 @@ def parseGFF3(gff3_path):
         ("parent", "Parent"),
         ("ID", "ID"),
     ]:
-        ft_df[cname] = (
-            ft_df["attributes"]
-            .str.split(attrid + "=", n=1, expand=True)[1]
-            .str.split(";", n=1, expand=True)[0]
-        )
+        try:
 
-    ft_df["GeneID"] = ft_df["attributes"].str.extract(r"Dbxref=.?GeneID:([\d]+)")[
-        0
-    ]  # better extract from within the Dbxref= attribute ?
+            ft_df[cname] = (
+                ft_df["attributes"]
+                .str.split(attrid + "=", n=1, expand=True)[1]
+                .str.split(";", n=1, expand=True)[0]
+            )
+        except:
+            logging.warning("An error occured while parsing the subattribute {} from the gff file. Probably absent.".format(attrid))
+            ft_df[cname] = float("NaN")
+
+    ft_df["GeneID"] = ft_df["attributes"].str.extract(r"Dbxref=.*?GeneID:([\d]+)")[0]
 
     ft_df.loc[
         (ft_df["feature"] == "CDS") & (ft_df["attributes"].str.contains("protein_id=")),
@@ -717,6 +748,24 @@ def profile_genome(t, window_ogs, ogtable, center, extension_size):
 
     return g
 
+def compare_taxon(t, window_ogs, ogtable, center, extension_size, label_map ):
+    logging.info("[{}] profiling ...".format(t))
+    g = profile_genome(t, window_ogs, ogtable, center, extension_size)
+    if g:
+        # runprofile.enable()
+        str_out = g.produceCMDLOutput(label_map=label_map)
+        # res = fg.WHITE + bg.BLACK + "IHateThis" + Style.RESET_ALL
+        # runprofile.disable()
+        # stats = pstats.Stats(runprofile).sort_stats('cumtime')
+        # stats.print_stats(10)
+    if str_out:
+        print(str_out)
+        logging.info("[{}] ... OK.".format(t))
+        return str_out
+    else:
+        logging.info("[{}] ... Failed.".format(t))
+    gc.collect()
+
 
 def main():
 
@@ -731,9 +780,9 @@ def main():
     if args.output_prefix:
         PREFIX = args.output_prefix
     else:
-        PREFIX = "prot_{}_k_{}".format(args.centerprotein_accession, args.k)
+        PREFIX = "prot_{}_ext_{}".format(args.centerprotein_accession, args.k)
 
-    logfilepath = pathlib.Path(args.outdir) / PREFIX + ".vicinator.log"
+    logfilepath = pathlib.Path(args.outdir) / "{}.vicinator.log".format(PREFIX)
     # logFormatter = logging.Formatter("%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
     logging.basicConfig(  # filename=str(logfilepath),
         level=logging.INFO,
@@ -775,7 +824,8 @@ def main():
     if len(center_index_list) > 1:
         logging.info(
             "centerprotein {} has multiple hits. Only the first hit is considered.".format(
-                args.centerprotein_accession)
+                args.centerprotein_accession
+            )
         )
     elif len(center_index_list) < 1:
         logging.exception(
@@ -806,6 +856,7 @@ def main():
 
     window_ogs = []
     for i, accession_set in enumerate(prot_accessions):
+        #print(i, accession_set)
         ogs = set()
         for acc in accession_set:
             og = ref_g.getOGidFromTable(ogtable, ref_g.name, acc)
@@ -887,31 +938,40 @@ def main():
 
     ####
     ####
-    full_output = []
     converter = Ansi2HTMLConverter(dark_bg=False, scheme="solarized", markup_lines=True)
 
-    for t in taxon_feature_files:
-        logging.info("[{}] profiling ...".format(t))
+    pool = mp.Pool(processes=args.nprocs)
+    manager = mp.Manager()
+    return_dict = manager.dict()
+    logging.info("Number of CPUs used: {}".format(args.nprocs))
+    full_output = pool.starmap(compare_taxon, [(t, window_ogs, ogtable, center, extension_size, label_map) for t in taxon_feature_files])
 
-        g = profile_genome(t, window_ogs, ogtable, center, extension_size)
-
-        if g:
-            # runprofile.enable()
-            str_out = g.produceCMDLOutput(label_map=label_map)
-            # res = fg.WHITE + bg.BLACK + "IHateThis" + Style.RESET_ALL
-            # runprofile.disable()
-            # stats = pstats.Stats(runprofile).sort_stats('cumtime')
-            # stats.print_stats(10)
-        if str_out:
-            print(str_out)
-            full_output.append(str_out)
-            logging.info("[{}] ... OK.".format(t))
-        else:
-            logging.info("[{}] ... Failed.".format(t))
-        gc.collect()
+    # full_output = []
+    #
+    # for t in taxon_feature_files:
+    #     logging.info("[{}] profiling ...".format(t))
+    #
+    #     g = profile_genome(t, window_ogs, ogtable, center, extension_size)
+    #
+    #     if g:
+    #         # runprofile.enable()
+    #         str_out = g.produceCMDLOutput(label_map=label_map)
+    #         # res = fg.WHITE + bg.BLACK + "IHateThis" + Style.RESET_ALL
+    #         # runprofile.disable()
+    #         # stats = pstats.Stats(runprofile).sort_stats('cumtime')
+    #         # stats.print_stats(10)
+    #     if str_out:
+    #         print(str_out)
+    #         full_output.append(str_out)
+    #         logging.info("[{}] ... OK.".format(t))
+    #     else:
+    #         logging.info("[{}] ... Failed.".format(t))
+    #     gc.collect()
 
     full_output = "\n".join(full_output)
-    htmloutputfilepath = pathlib.Path(args.outdir) / PREFIX + "vicinator.out.html"
+    htmloutputfilepath = pathlib.Path(args.outdir) / "{}.vicinator.out.html".format(
+        PREFIX
+    )
     with open(htmloutputfilepath, "w") as htmlout:
         htmlout.write(converter.convert(full_output))
 
